#include "stdafx.h"

#include <assert.h>
#include "CaseServer.h"
#include "CasePackage.h"
#include "CCaseManagment.h"

#include <fstream>
using namespace std;

#pragma warning (disable : 4996)

bool CaseServer::m_stopAction = false;
akl_string CaseServer::m_userName = "";
akl_string CaseServer::m_password = "";
 
CaseServer::CaseServer(akl_string caseApiUrl)
    : m_caseApiUrl(caseApiUrl),
      m_isError(false), m_isConnected(false),
      m_hSession(0), m_hConnection(0), m_hRequest(0),
      m_errorCode(0), m_errorString(""),
      m_isUseProxy(false), m_proxyUrl(""), m_proxyUser(""), m_proxyPass(""), 
      m_isUseSsl(false), m_contentLength(0)
{   
}

CaseServer::~CaseServer()
{
    CloseConnection();
}
 
void CaseServer::UseProxy(akl_string proxyUrl, akl_string proxyUser, akl_string proxyPass)
{
    if(proxyUrl.length())
    {
        m_isUseProxy = true;
        m_proxyUrl   = proxyUrl;
        m_proxyUser  = proxyUser;
        m_proxyPass  = proxyPass;
    }
    else
    {
        m_isUseProxy = false;
        m_proxyUrl   = "";
        m_proxyUser  = "";
        m_proxyPass  = "";
    }
}
 
bool CaseServer::Authenticate(akl_string userName, akl_string password)
{
    // Set user name and password 
    CaseServer::m_userName = userName;
    CaseServer::m_password = password;

    akl_string requestUrl = _T("/server_url");
    if(CreateRequest(requestUrl))
    {
        DWORD status = GetResponse();
        if(status == HTTP_STATUS_OK)
        {  
            CloseConnection();
            return true;
        }
        else
        {
            SetHttpError(status);
        }
    }

    CaseServer::m_userName = "";
    CaseServer::m_password = "";
    CloseConnection();
    return false;
}

akl_string CaseServer::DownloadCaseList(bool toBeMarked)
{
    akl_string responseData = "";
    akl_string caseListUrl = (toBeMarked) ?  
        _T("/url/to_be_marked") : _T("/url");
    
    if(CreateRequest(caseListUrl))
    {
        // Get response from server.
        DWORD responseResult = GetResponse();
        if(responseResult == HTTP_STATUS_OK)
        {   
            char* pBuffer = 0L;
            int readCount = ReadResponse(&pBuffer);

            // Download case list xml data.
            while(readCount > 0)
            {
                responseData += akl_string(pBuffer);
                delete[] pBuffer;
                readCount = ReadResponse(&pBuffer);

                // Check for possible errors.
                if(readCount < 0)
                {
                    return "";
                }
            }
        }
        else
        {
            SetHttpError(responseResult);
        }
    }

    CloseConnection();
    return responseData;
}
bool CaseServer::DownloadCase(struct _casem& aCase, 
                              CProgressCtrl* downloadBar, 
                              CStatic* labelStatus)
{   
    // Start downloading.
    // Set case API url to provided with case.
    m_caseApiUrl = FixUrl(aCase.caseData);
    if(CreateRequest(aCase.caseData))
    {
        // Get response from server.
        DWORD responseResult = GetResponse();
        if(responseResult == HTTP_STATUS_OK)
        {
            m_totalDownloaded = 0;
            m_contentLength = GetHeaderNumberValue(eContentLength) / 1024;

            ofstream outFile;
            CasePackage package(aCase);
            
            // Download case list xml data.
            char* pBuffer = 0L;
            int readCount = ReadResponse(&pBuffer);
            CaseServer::m_stopAction = false;
            while(readCount > 0)
            {
                // Open file if needed.
                if(!outFile.is_open())
                {   
                    int result = SHCreateDirectoryEx(NULL, package.GetPath(), NULL);
                    outFile.open(package.GetPathName(), ios::binary);
                    if(!outFile.is_open())
                    { 
                        akl_string errorString;
                        errorString.format(L"Failed to open file to download the case. SHCreateDirectoryEx result code: %d", result);
                        SetError(errorString, GetLastError());
                        return false;
                    }
                }
                // write to file
                outFile.write(pBuffer, readCount);
                outFile.flush();
                delete[] pBuffer;

                UpdateDownloadUI(downloadBar, labelStatus, readCount);
                
                // Read and check for possible errors.
                readCount = ReadResponse(&pBuffer);
                if(readCount < 0 || CaseServer::m_stopAction)
                {
                    outFile.close();
                    ::DeleteFile(package.GetPathName());                     
                    if(CaseServer::m_stopAction)
                    {
                        // when user cancel it is not download fault; so return true.
                        SetError("User cancels downloading...");
                        return true;
                    }
                    return false;
                }
            }
            outFile.close();            
            package.ExtractAll(package.GetPathName());
            return true;
        }
        else
        {
            SetHttpError(responseResult);
        }
    }
    return false;
}

bool CaseServer::UploadCase(struct _casem& aCase)
{   
    CasePackage package(aCase);

    // Prepare package for transmittion
    if (!package.OpenRaw() || !package.ResetRaw())
    {
        SetError("Can't access uploaded package.", GetLastError());
        return false;
    }

    // Create post request.
    // Set case API url to provided with case.
    m_caseApiUrl = FixUrl(aCase.marginCurveURL);
    if(CreateRequest(aCase.marginCurveURL, ePost))
    {
        // Constructing multipart request header
        akl_string contentType = "Content-Type: multipart/form-data; boundary=19024605111143684786787635207";
        akl_string multipartRequestTeml =  "--19024605111143684786787635207\r\n"
            "Content-Disposition: form-data; name=\"file\"; filename=\"%s\"\r\n"
            "Content-Type: application/octet-stream\r\n\r\n";

        const char* multipartPost = "\r\n--19024605111143684786787635207--\r\n";

        // Create request
        akl_string request = "";
        request.format(multipartRequestTeml, package.GetName());
        int requestPreLength = request.length() + 1;
        char* multipartRequestPre = new char[requestPreLength];
        wcstombs(multipartRequestPre, request, requestPreLength);
       
        //Send multipart request header        
        int totalLength = requestPreLength + package.GetSize() + strlen(multipartPost);
        if(!WinHttpSendRequest(m_hRequest, contentType, contentType.length(), 
            (LPVOID*)multipartRequestPre, requestPreLength, totalLength, 0))
        {
            SetError("Error in WinHttpSendRequest for multipart request.", GetLastError());
            return false;
        }
        delete [] multipartRequestPre;

        //Send file        
        DWORD dwNumberOfBytes = 0;
        CaseServer::m_stopAction = false;
        int readCount = package.ReadRaw();
        while(readCount > 0)
        {               
            if (package.IsError())
            {   
                SetError(_T("Error while read uploaded package file."));                
                return false;
            }
            
            if (!WinHttpWriteData(m_hRequest, package.GetReadBuffer(), 
                                  readCount, &dwNumberOfBytes))
            {
                SetError("Error in WinHttpWriteData when uploading case.", GetLastError());
                return false;
            }
            // Read next portion of bytes.
            readCount = package.ReadRaw();
            if(CaseServer::m_stopAction)
            {
                package.CloseRaw();
                ::DeleteFile(package.GetPathName());
                SetError("User cancels uploading.");
                return true;
            }
        }

        // Send request post
        if (!WinHttpWriteData(m_hRequest, multipartPost, 
                              (DWORD)strlen(multipartPost), &dwNumberOfBytes))
        {
            SetError("Error in WinHttpWriteData when uploading package.", GetLastError());
            return false;
        }        
        // Delete package file.
        package.CloseRaw();
        ::DeleteFile(package.GetPathName());
        // Check response errors
        if(!WinHttpReceiveResponse(m_hRequest, NULL))
        {   
            SetError("Error in WinHttpReceiveResponse when uploading package.", GetLastError());
            return false;
        }

        // Check status code.
        DWORD statusCode = GetHeaderNumberValue(eStatusCode);        
        if(statusCode != HTTP_STATUS_CREATED)
        {
            SetHttpError(statusCode);
            return false;
        }
        CloseConnection();
        return true;
    }
    return false;
}

bool CaseServer::UploadArticulator(akl_string caseNumber, akl_string articulatorType)
{   
    akl_string resourceUrl = akl_string("/vamm/") + caseNumber;

    if(CreateRequest(resourceUrl, ePut))
    {
        // Constructing multipart request header
        akl_string contentType = "Content-Type: multipart/form-data; boundary=19024605111143684786787635207";
        akl_string multipartRequestTempl = "--19024605111143684786787635207\r\n"
            "Content-Disposition: form-data; name=\"device[articulator]\"\r\n\r\n"
            "%s\r\n--19024605111143684786787635207--\r\n";
        
        // Create request
        akl_string request = "";
        request.format(multipartRequestTempl, articulatorType);
        
        int length = request.length() + 1;
        char* multipartRequest = new char[length];
        wcstombs(multipartRequest, request, length);
               
        //Send multipart request header
        if(!WinHttpSendRequest(m_hRequest, contentType, contentType.length(), 
            (LPVOID*)multipartRequest, length, length, 0))
        {
            SetError("Error in WinHttpSendRequest for multipart request.", GetLastError());
            return false;
        }
        delete [] multipartRequest;

        // Recive response.
        if (!WinHttpReceiveResponse(m_hRequest, NULL))
        {   
            SetError("Error in WinHttpReceiveResponse when uploading articulator.", GetLastError());
            return false;
        }

        // Check status code.
        DWORD statusCode = GetHeaderNumberValue(eStatusCode);        
        if(statusCode != HTTP_STATUS_OK)
        {
            SetHttpError(statusCode);           
            return false;
        }
        CloseConnection();
        return true;        
    }
    return false;
}
// network iteraction logic 
bool CaseServer::CreateRequest(akl_string requestUrl, RequestType type)
{
    // New request must close any existing connection.
    CloseConnection();    
    if(!OpenConnection())
    {
        return false;
    }
    // Create request.
    if(type == eGet)
    {
        const wchar_t* accept_type [] = { _T("application/xml"), NULL };
        m_hRequest = WinHttpOpenRequest(m_hConnection, L"GET", FixResourceUrl(requestUrl),
                                         NULL, WINHTTP_NO_REFERER, accept_type,
                                         m_isUseSsl ? WINHTTP_FLAG_SECURE : 0);
    }
    else if(type == ePost)
    {
        const wchar_t* accept_type [] = { _T("*/*"), NULL };
        m_hRequest = WinHttpOpenRequest(m_hConnection, L"POST", FixResourceUrl(requestUrl),
                                        NULL, WINHTTP_NO_REFERER, accept_type,
                                        m_isUseSsl ? WINHTTP_FLAG_SECURE : 0);
    }
    else if(type == ePut)
    {
        const wchar_t* accept_type [] = { _T("*/*"), NULL };
        m_hRequest = WinHttpOpenRequest(m_hConnection, L"PUT", FixResourceUrl(requestUrl),
                                        NULL, WINHTTP_NO_REFERER, accept_type,
                                        m_isUseSsl ? WINHTTP_FLAG_SECURE : 0);
    }
    else
    {
        SetError("Undefined request type in CreateRequest.");
        return false;
    }
    if(!m_hRequest)
    {
        SetError("Can't create request!...");
        return false;
    }

    // Set credentials
    if(!SetCredentials())
    {
        return false;
    }

    m_isConnected = true;
    return true;
}

DWORD CaseServer::GetResponse()
{   
    if(m_isConnected)
    {        
        BOOL result = WinHttpSendRequest(m_hRequest, 
                                         WINHTTP_NO_ADDITIONAL_HEADERS, 0, 
                                         WINHTTP_NO_REQUEST_DATA, 0, 0, 0);
        if (result)
        {
            result = WinHttpReceiveResponse(m_hRequest, NULL);
        }
        else
        {
            SetError("Error while recive response!", GetLastError());
        }
        if (result)
        {
            return GetHeaderNumberValue(eStatusCode);
        }
    }          
    return -1;
}

int CaseServer::ReadResponse(char** ppBuffer)
{
    *ppBuffer = 0L;
    // Get size of available data
    DWORD dwSize = 0;
    if(!WinHttpQueryDataAvailable(m_hRequest, &dwSize))
    {
        SetError("Error in WinHttpQueryDataAvailable.", GetLastError());
        CloseConnection();
        return -1;
    }
    // If available data grater than 0.
    if(dwSize)
    {
        // Allocate space for the buffer.
        *ppBuffer = new char[dwSize+1];                
        ZeroMemory(*ppBuffer, dwSize+1);
        
        // Read data from response
        DWORD dwDownloaded = 0;
        if(!WinHttpReadData(m_hRequest, (LPVOID)(*ppBuffer), dwSize, &dwDownloaded))
        {
            delete [] *ppBuffer;
            *ppBuffer = 0L;
            SetError("Error in WinHttpReadData.", GetLastError());
            CloseConnection();
            return -1;
        }
        m_totalDownloaded += dwDownloaded;
    }
    return (int)dwSize;
}
bool CaseServer::OpenConnection()
{    
    // Choose port
    INTERNET_PORT nPort = (m_isUseSsl) ? 
        INTERNET_DEFAULT_HTTPS_PORT : INTERNET_DEFAULT_HTTP_PORT;

    // Create session first
    if (m_isUseProxy)
    {
        m_hSession = WinHttpOpen(GetVammAgent(), 
                                 WINHTTP_ACCESS_TYPE_NAMED_PROXY, 
                                 FixUrl(m_proxyUrl), NULL, 0);
    }
    else
    {
        m_hSession = WinHttpOpen(GetVammAgent(), 
                                 WINHTTP_ACCESS_TYPE_NO_PROXY, 
                                 WINHTTP_NO_PROXY_NAME, NULL, 0);
    }    
    if(!m_hSession)
    {
        SetError("Can't create session!...", GetLastError());
        return false;
    }

    // Set new timeout value
    DWORD dwData = 0;
    DWORD dwSize = sizeof(DWORD);
    /*WinHttpQueryOption( m_hSession, WINHTTP_OPTION_CONNECT_TIMEOUT, &dwData, &dwSize);
    TRACE("Old timeout is %d\n", dwData);*/
    dwData = 300000;

    WinHttpSetOption( m_hSession, WINHTTP_OPTION_CONNECT_TIMEOUT, &dwData, dwSize );
    TRACE("New timeout is %d\n", dwData);
    
    // Opens connection
    m_hConnection = WinHttpConnect(m_hSession, FixUrl(m_caseApiUrl), nPort, 0);
    if(!m_hConnection)
    {
        SetError("Can't open connection!...", GetLastError());
        return false;    
    }
    return true;
}

bool CaseServer::SetCredentials()
{   
    // If Proxy is used set proxy credentials
    if (m_isUseProxy)
    {
        if(!WinHttpSetCredentials(m_hRequest, 
                               WINHTTP_AUTH_TARGET_PROXY,
                               WINHTTP_AUTH_SCHEME_BASIC, 
                               m_proxyUser, m_proxyPass, NULL))
        {
            SetError("Proxy user name or password are invalid!...", GetLastError());
            return false;
        }
    }
     
    if(!WinHttpSetCredentials(m_hRequest, 
                              WINHTTP_AUTH_TARGET_SERVER,
                              WINHTTP_AUTH_SCHEME_BASIC, 
                              CaseServer::m_userName, 
                              CaseServer::m_password, NULL))
     {
         SetError("User name or password are invalid!...", GetLastError());
         return false;
     }
     return true;
}

void CaseServer::CloseConnection()
{
    if (m_hRequest)
    {
        WinHttpCloseHandle(m_hRequest);
        m_hRequest = 0;
    }

    if (m_hConnection)
    {
        WinHttpCloseHandle(m_hConnection);
        m_hConnection = 0;
    }

    if (m_hSession)
    {
        WinHttpCloseHandle(m_hSession);
        m_hSession = 0;
    }

    m_contentLength = -1;
    m_isConnected = false;
    m_isError = false;
}

// utility methods
DWORD CaseServer::GetHeaderNumberValue(HeaderNumberValue header)
{
    DWORD value = -1;
    DWORD valueSize = sizeof(value);
    if(eStatusCode == header || eContentLength == header)
    {
        DWORD requestedHeader = (eStatusCode == header) ? 
            WINHTTP_QUERY_STATUS_CODE : WINHTTP_QUERY_CONTENT_LENGTH;

        WinHttpQueryHeaders(m_hRequest, 
            requestedHeader | WINHTTP_QUERY_FLAG_NUMBER, 
            NULL, &value, &valueSize, NULL);
    }
    return value;
}

akl_string CaseServer::FixUrl(akl_string url)
{
    // Fix url so that "https://test.com/something/somethingelse" becomes "test.com"
    akl_string result = url;

    int leftBound = url.find("://");
    if(leftBound > -1)
    {
        result = result.right(result.length() - leftBound - 3);

        int rightBound = result.find("/");
        if (rightBound > -1)
        {
            result = result.left(rightBound);
        }
    }
    return result;
}
akl_string CaseServer::FixResourceUrl(akl_string url)
{
    akl_string result = url;
    int startPos = result.find(_T("://"));
    if(startPos > -1)
    {
        startPos += 3;        
        result = result.right(url.length() - startPos);
        startPos = result.find('/');
        result = result.right(result.length() - startPos);
    }
    if(result.left(1) != "/")
    {
        result.prepend("/");
    }
    return result;
}
void CaseServer::SetError(akl_string errorString, int errorCode)
{
    m_isError = true;
    m_errorString = errorString;
    m_errorCode = errorCode;
}

void CaseServer::SetHttpError(int errorCode)
{
    m_isError = true;
    m_errorCode = errorCode;
    switch(errorCode)
    {
        case 0: //Server connection error
            m_errorString = _T("Error: Please check/set connection preferences under: Tools->Preferences menu.");
            break;

        case 301: //Moved Permanently
            m_errorString = _T("Error: Requested document moved permanently!");
            break;

        case 400: //Bad Request
            m_errorString = _T("Error: The request could not be understood by the server due to malformed syntax!");
            break;

        case 401: //Server auth requiested
            m_errorString = _T("Error: Invalid server username or password!");
            break;

        case 403: //Forbidden
            m_errorString = _T("Error: The server understood the request, but is refusing to fulfill it!");
            break;

        case 404: //Not found
            m_errorString = _T("Error: Case data was not found on server or is not attached!");
            break;

        case 406: //Internal Server Error
            m_errorString = _T("Error: Internal Server Error!");
            break;

        case 407: //Proxy auth requested
            m_errorString = _T("Error: Invalid proxy username or password!");
            break;

        case 500: //Not Acceptable
            m_errorString = _T("Error: The server does not support the functionality required to fulfill the request!");
            break;

        case 504: // Gateway Timeout 
            m_errorString = _T("Error: The server, while acting as a gateway or proxy, did not receive a timely response from the upstream server!");
            break;
        default:
            m_errorString.format(_T("Error: Server returned error code (%d)"), errorCode);
    }
}

akl_string CaseServer::GetVammAgent()
{
    akl_string version = theApp.GetAppVersionString();
    int pos = version.reverse_find('.');

    assert(pos != -1);
    if (pos == -1)
    {
        akl_string agent = L"Mozilla-FireFox";
        return agent;
    }

    version[pos] = '-';
    version.insert('r', pos + 1);
    akl_string agent = L"Mozilla-FireFox";
    return agent;
}

// Update UI logic
void CaseServer::UpdateDownloadUI(CProgressCtrl* downloadBar, CStatic* labelStatus, int readBytes)
{   
    int pos = (m_totalDownloaded / 1024) / (m_contentLength / 100 + 0.00001f);
    downloadBar->SetPos(pos);
    
    akl_string message;
    message.format(_T("Downloading: %d kb of %.0f kb is done."), 
        m_totalDownloaded / 1024, (float)m_contentLength);
    
    labelStatus->SetWindowText(message);
    TRACE("downloading data %d done !\n", pos);

    // Update dialog and MainFrame
    HWND hDialog = downloadBar->GetParent()->m_hWnd;
    HWND hMainFrame = AfxGetMainWnd()->m_hWnd;

    MSG msg;
    while (PeekMessage(&msg, hMainFrame, 0, 0, PM_REMOVE))
    {
        DispatchMessage(&msg);
        UpdateWindow(hMainFrame);
    }

    while (PeekMessage(&msg, hDialog, 0, 0, PM_REMOVE))
    {
        DispatchMessage(&msg);
        UpdateWindow(hDialog);
    }
}
